/*
Szymon Rusinkiewicz
Princeton University

mesh_shade.cc
Apply procedural shaders to a mesh
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>
#include <algorithm>
#include "TriMesh.h"
#include "TriMesh_algo.h"
#include "KDtree.h"
using namespace std;


#ifdef WIN32
# ifndef strncasecmp
#  define strncasecmp strnicmp
# endif
#endif


// Quick 'n dirty portable random number generator 
static inline float tinyrnd()
{
	static unsigned trand = 0;
	trand = 1664525u * trand + 1013904223u;
	return (float) trand / 4294967296.0f;
}


// Apply a solid color to the mesh
void solidcolor(TriMesh *mesh, const char *col)
{
	unsigned c;
	sscanf(col, "%x", &c);
	int r = (c >> 16) & 0xff;
	int g = (c >> 8)  & 0xff;
	int b =  c        & 0xff;
	Color cc = Color(r,g,b);
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++)
		mesh->colors[i] = cc;
}


// Color based on normals
void colorbynormals(TriMesh *mesh)
{
	mesh->need_normals();
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++) {
		const vec &n = mesh->normals[i];
		int r = int(min(max(128.0*(n[0]+1.0), 0.0), 255.99));
		int g = int(min(max(128.0*(n[1]+1.0), 0.0), 255.99));
		int b = int(min(max(128.0*(n[2]+1.0), 0.0), 255.99));
		mesh->colors[i] = Color(r,g,b);
	}
}



// Compute a "feature size" for the mesh: computed as 1% of 
// the reciprocal of the 10-th percentile curvature 
float feature_size(TriMesh *mesh)
{
	mesh->need_curvatures();
	int nv = mesh->curv1.size();
	int nsamp = min(nv, 500);
                
	vector<float> samples;
	samples.reserve(nsamp * 2);
        
	for (int i = 0; i < nsamp; i++) {
		// Quick 'n dirty portable random number generator  
		static unsigned randq = 0;
		randq = unsigned(1664525) * randq + unsigned(1013904223);

		int ind = int(tinyrnd() * nv);
		samples.push_back(fabs(mesh->curv1[ind]));
		samples.push_back(fabs(mesh->curv2[ind]));
	}
	const float frac = 0.1f;
	const float mult = 0.01f;
	int which = int(frac * samples.size());
	nth_element(samples.begin(), samples.begin() + which, samples.end());
	float f = mult / samples[which];
	fprintf(stderr, "Feature size = %f\n", f);
	return f;
}


// Color based on curvature
void colorbycurv(TriMesh *mesh, const char *scale, const char *smooth)
{
	mesh->need_curvatures();
	float smoothsigma = atof(smooth);
	if (smoothsigma > 0.0f) {
		smoothsigma *= mesh->feature_size();
		diffuse_curv(mesh, smoothsigma);
	}
	float cscale = sqr(10.0f * atof(scale) * feature_size(mesh));

	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++) {
		float H = 0.5f * (mesh->curv1[i] + mesh->curv2[i]);
		float K = mesh->curv1[i] * mesh->curv2[i];
		float h = 4.0f / 3.0f * fabs(atan2(H*H-K,H*H*sgn(H)));
		float s = M_2_PI * atan((2.0f*H*H-K)*cscale);
		mesh->colors[i] = Color::hsv(h,s,1.0);
	}
}


// Color based on curvature.  Similar to above, but uses a grayscale mapping.
void gcolorbycurv(TriMesh *mesh, const char *scale, const char *smooth)
{
	mesh->need_curvatures();
	float smoothsigma = atof(smooth);
	if (smoothsigma > 0.0f) {
		smoothsigma *= mesh->feature_size();
		diffuse_curv(mesh, smoothsigma);
	}
	float cscale = 10.0f * atof(scale) * feature_size(mesh);

	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++) {
		float H = 0.5f * (mesh->curv1[i] + mesh->curv2[i]);
		int c = int(min(max(256.0 / M_PI * (atan(H*cscale) + M_PI_2),
				    0.0), 255.99));
		mesh->colors[i] = Color(c,c,c);
	}
}


// Accessibility shading
void acc(TriMesh *mesh, const char *maxsize_, const char *offset_)
{
	mesh->need_normals();
	float fs = feature_size(mesh);
	float maxsize = atof(maxsize_) * fs;
	float offset = atof(offset_) * fs;

	KDtree *kd = new KDtree(mesh->vertices);
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++) {
		const vec &n = mesh->normals[i];
		point p = mesh->vertices[i] + offset * n;
		float tmin = 0, tmax = maxsize;
		for (int iter = 0; iter < 8; iter++) {
			float tmid = 0.5f * (tmin + tmax);
			point q = p + tmid * n;
			const float *qq = kd->closest_to_pt(q, sqr(tmid));
			if (qq)
				tmax = tmid;
			else
				tmin = tmid;
		}
		int c = int(min(128.0 * (tmin + tmax) / maxsize, 255.99));
		mesh->colors[i] = Color(c,c,c);
	}
}


// Color by distance to bdy
void bdyshade(TriMesh *mesh, const char *nedges_)
{
	int nedges = atoi(nedges_);
	int nv = mesh->vertices.size();
	mesh->need_neighbors();
	mesh->flags.resize(nv);
	for (int i = 0; i < nv; i++)
		mesh->flags[i] = mesh->is_bdy(i) ? 0 : nedges;
	for (int iter = 1; iter < nedges; iter++) {
		for (int i = 0; i < nv; i++) {
			for (int j = 0; j < mesh->neighbors[i].size(); j++) {
				int n = mesh->neighbors[i][j];
				if (mesh->flags[n] + 1 < mesh->flags[i])
					mesh->flags[i] = mesh->flags[n] + 1;
			}
		}
	}
	for (int i = 0; i < nv; i++) {
		int c = int(255.99f * mesh->flags[i] / nedges);
		mesh->colors[i] = Color(255, c, c);
	}
}


void usage(const char *myname)
{
	fprintf(stderr, "Usage: %s model shader [options] outfile\n", myname);
	fprintf(stderr, "Shaders:\n");
	fprintf(stderr, "	color rrggbb	Solid color (specified in hex)\n");
	fprintf(stderr, "	normals		Colored based on normals (r = nx, etc.)\n");
	fprintf(stderr, "	curv sc sm	Colored based on curvature (args = scale, smoothing)\n");
	fprintf(stderr, "	gcurv sc sm	Grayscale based on curvature (args = scale, smoothing)\n");
	fprintf(stderr, "	acc max off	Accessibility (args = maximum size, offset)\n");
	fprintf(stderr, "	bdy max		Distance to boundary (arg = maximum # edges)\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	if (argc < 4)
		usage(argv[0]);
	const char *infilename = argv[1];

	TriMesh *themesh = TriMesh::read(infilename);
	if (!themesh)
		usage(argv[0]);
	themesh->colors.resize(themesh->vertices.size());

	const char *shader = argv[2];
	const char *outfilename = argv[3];
	if (!strncasecmp(shader, "col", 3)) {
		if (argc < 5) {
			fprintf(stderr, "\n\"color\" needs one argument\n\n");
			usage(argv[0]);
		}
		solidcolor(themesh, argv[3]);
		outfilename = argv[4];
	} else if (!strncasecmp(shader, "norm", 4)) {
		colorbynormals(themesh);
	} else if (!strncasecmp(shader, "curv", 4)) {
		if (argc < 6) {
			fprintf(stderr, "\n\"curv\" needs two arguments\n\n");
			usage(argv[0]);
		}
		colorbycurv(themesh, argv[3], argv[4]);
		outfilename = argv[5];
	} else if (!strncasecmp(shader, "gcurv", 5)) {
		if (argc < 6) {
			fprintf(stderr, "\n\"gcurv\" needs two arguments\n\n");
			usage(argv[0]);
		}
		gcolorbycurv(themesh, argv[3], argv[4]);
		outfilename = argv[5];
	} else if (!strncasecmp(shader, "acc", 3)) {
		if (argc < 6) {
			fprintf(stderr, "\n\"acc\" needs two arguments\n\n");
			usage(argv[0]);
		}
		acc(themesh, argv[3], argv[4]);
		outfilename = argv[5];
	} else if (!strncasecmp(shader, "bdy", 3)) {
		if (argc < 5) {
			fprintf(stderr, "\n\"bdy\" needs one argument\n\n");
			usage(argv[0]);
		}
		bdyshade(themesh, argv[3]);
		outfilename = argv[4];
	} else {
		fprintf(stderr,"\nUnknown shader [%s]\n\n", shader);
		usage(argv[0]);
	}

	themesh->write(outfilename);
}


